;*=====================================================================*/
;*    serrano/prgm/project/skribe/tools/skribeinfo/engine.scm          */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Sun Oct 14 11:35:59 2001                          */
;*    Last change :  Mon Sep 15 10:07:02 2003 (serrano)                */
;*    Copyright   :  2001-03 Manuel Serrano                            */
;*    -------------------------------------------------------------    */
;*    The construction of the form that is a valid Skribe tree.        */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __skribeinfo_engine
   (import __skribeinfo_param)
   (export *indexes*
	   (bind-variable ::bstring ::obj)
	   (variable-bound?::bool ::bstring)
	   (syncodeindex ::symbol ::symbol)
	   (synindex ::symbol ::symbol)
	   (texi->skribe ::obj)
	   (printindex ::symbol)
	   (index-entry ::symbol ::obj)))

;*---------------------------------------------------------------------*/
;*    Indexes                                                          */
;*---------------------------------------------------------------------*/
(define *indexes* '(*cp* *fn* *vr* *ky* *pg* *tp*))

(define *merged-indexes*
   (map (lambda (cell) (list cell cell)) *indexes*))

;*---------------------------------------------------------------------*/
;*    printindex ...                                                   */
;*---------------------------------------------------------------------*/
(define (printindex kind)
   (let ((indexes (assq kind *merged-indexes*)))
      (if (pair? indexes)
	  `(the-index :column *skribeinfo-index-column* ,@(cdr indexes))
	  (error "printindex" "Illegal index" kind))))

;*---------------------------------------------------------------------*/
;*    index-entry ...                                                  */
;*---------------------------------------------------------------------*/
(define (index-entry index val)
   (cond
      ((string? val)
       `(index :index ,index ,val))
      (else
       (match-case val
	  (((? string?))
	   `(index :index ,index ,(car val)))
	  (((?opt (and ?str (? string?))))
	   `(index :index ,index ,str :shape ,@val))
	  ((?opt (and ?str (? string?)))
	   `(index :index ,index ,str :shape ,val))
	  (((?opt (and ?str1 (? string?)) (and ?str2 (? string?))))
	   `(index :index ,index ,(string-append str1 str2) :shape ,@val))
	  ((?opt (and ?str1 (? string?)) (and ?str2 (? string?)))
	   `(index :index ,index ,(string-append str1 str2) :shape ,val))
	  (else
	   (error "index-entry"
		  "Illegal index entry (should be a string)"
		  val))))))
	   
;*---------------------------------------------------------------------*/
;*    merge-index ...                                                  */
;*---------------------------------------------------------------------*/
(define (merge-index l1 l2 fun)
   (cond
      ((not (memq l2 *indexes*))
       (error fun "Illegal index" l2))
      ((not (memq l1 *indexes*))
       (error fun "Illegal index" l1))
      (else
       (let ((cell (assq l2 *merged-indexes*)))
	  (set-cdr! cell (cons l1 (cdr cell)))))))

;*---------------------------------------------------------------------*/
;*    syncodeindex ...                                                 */
;*---------------------------------------------------------------------*/
(define (syncodeindex l1 l2)
   (merge-index l1 l2 'syncodeindex))

;*---------------------------------------------------------------------*/
;*    synindex ...                                                     */
;*---------------------------------------------------------------------*/
(define (synindex l1 l2)
   (merge-index l1 l2 'synindex))

;*---------------------------------------------------------------------*/
;*    *variables* ...                                                  */
;*---------------------------------------------------------------------*/
(define *variables* (make-hashtable))

;*---------------------------------------------------------------------*/
;*    bind-variable ...                                                */
;*---------------------------------------------------------------------*/
(define (bind-variable var val)
   (begin
      (hashtable-put! *variables* var val)
      ""))

;*---------------------------------------------------------------------*/
;*    variable-bound? ...                                              */
;*---------------------------------------------------------------------*/
(define (variable-bound? var)
   (hashtable-get *variables* var))

;*---------------------------------------------------------------------*/
;*    texi-eval ...                                                    */
;*---------------------------------------------------------------------*/
(define (texi-eval sexp)
   (define (inner sexp)
      (if (not (pair? sexp))
	  sexp
	  (match-case sexp
	     ((value ?var)
	      (let ((val (hashtable-get *variables* var)))
		 (if (not val)
		     (error "texi-eval" "Can't find variable definition" var)
		     (inner `(list ,@val)))))
	     (else
	      (if (null? (car sexp))
		  (inner (cdr sexp))
		  (cons (inner (car sexp)) (inner (cdr sexp))))))))
   (if (null? sexp)
       ""
       (let ((res (inner sexp)))
	  (if (and (pair? res) (string? (car res)) (null? (cdr res)))
	      (car res)
	      res))))

;*---------------------------------------------------------------------*/
;*    texi->skribe ...                                                 */
;*    -------------------------------------------------------------    */
;*    This function accepts a flow of sexps that it constructs         */
;*    a correclty formed Skribe expression representing the whole      */
;*    document.                                                        */
;*    -------------------------------------------------------------    */
;*    This function implements a kind of automata that keeps track     */
;*    of what it is currently doing. For instance, it has to           */
;*    remember what is the current chapter, section, ... in order      */
;*    to accumulate the texts and give them to the proper owner.       */
;*    That is, we have to turn expressions such as:                    */
;*       (chapter "chap")                                              */
;*       (section "sec")                                               */
;*       "Foo bar"                                                     */
;*    into:                                                            */
;*       (chapter "chap"                                               */
;*        (section "sec"                                               */
;*         "Foo bar"))	                                               */
;*---------------------------------------------------------------------*/
(define (texi->skribe sexp)
   ;; first we have to find the title
   (seek-title sexp))

;*---------------------------------------------------------------------*/
;*    seek-title ...                                                   */
;*    -------------------------------------------------------------    */
;*    This function evaluates but does not include in the generated    */
;*    Skribe file everything it finds before the title.                */
;*---------------------------------------------------------------------*/
(define (seek-title sexp)
   (if (pair? sexp)
       (match-case (car sexp)
	  ((title . ?sexps)
	   (let* ((title (title->skribe sexps))
		  (body (body->skribe (cdr sexp))))
	      (cond
		 ((not (list? title))
		  (error "skribeinfo" "Illegal title" title))
		 ((not (list? body))
		  (error "skribeinfo" "Illegal body" body))
		 (else
		  (append title body)))))
	  (else
	   (texi-eval (car sexp))
	   (seek-title (cdr sexp))))
       #f))

;*---------------------------------------------------------------------*/
;*    title->skribe ...                                                */
;*---------------------------------------------------------------------*/
(define (title->skribe sexp)
   (let liip ((sexp sexp)
	      (titles '())
	      (authors '())
	      (blurp '()))
      (if (pair? sexp)
	  (match-case (car sexp)
	     (()
	      (liip (cdr sexp)
		    titles
		    authors
		    blurp))
	     ((title . ?title)
	      (liip (cdr sexp)
		    (cons (texi-eval title) titles)
		    authors
		    blurp))
	     ((subtitle . ?stitle)
	      (liip (cdr sexp)
		    (append (reverse (map texi-eval stitle))
			    (cons '(linebreak) titles))
		    authors
		    blurp))
	     ((author . ?author)
	      (liip (cdr sexp)
		    titles
		    (cons (texi-eval author) authors)
		    blurp))
	     (else
	      (liip (cdr sexp)
		    titles
		    authors
		    (cons (texi-eval (car sexp)) blurp))))
	  `(document :title (list ,@(reverse! titles))
		     :author (list ,@(map (lambda (a)
					     (let ((a (if (string? a)
							  (list a)
							  a)))
						`(author :name (list ,@a))))
					  (reverse! authors)))
		     ,@(if (pair? blurp)
			   `((list ,@(reverse! blurp)))
			   '())))))

;*---------------------------------------------------------------------*/
;*    body->skribe ...                                                 */
;*    -------------------------------------------------------------    */
;*    Read all the remaining expressions and store them into one       */
;*    list that will be the body of the document.                      */
;*---------------------------------------------------------------------*/
(define (body->skribe sexp)
   (let laa ((sexp sexp)
	      (res '()))
      (if (null? sexp)
	  res
	  (multiple-value-bind (skribe rest)
	     (sexp->skribe sexp '())
	     (laa rest (append skribe res))))))

;*---------------------------------------------------------------------*/
;*    sexp->skribe ...                                                 */
;*---------------------------------------------------------------------*/
(define (sexp->skribe sexp stop)
   (define (nesting block)
      (memq block '(paragraph subsubsection subsection section chapter)))
   (let luup ((sexp sexp)
	      (res '()))
      (if (null? sexp)
	  (values (reverse! res) '())
	  (let ((exp (car sexp))
		(rest (cdr sexp)))
	     (cond
		((or (string? exp) (number? exp))
		 (luup rest (cons exp res)))
		((or (eq? exp #unspecified) (eq? exp '()))
		 (luup rest res))
		((pair? exp)
		 (match-case exp
		    ((list . ?exps)
		     (luup (append exps rest) res))
		    (((and ?block (or paragraph subsubsection subsection section chapter)) . ?sec)
		     (if (memq block stop)
			 (values (reverse! res) sexp)
			 (multiple-value-bind (body rest)
			    (sexp->skribe rest (append (nesting block) stop))
			    (luup rest (cons (append exp body) res)))))
		    (else
		     (luup rest (cons (texi-eval exp) res)))))
		(else
		 (error "sexp->skribe" "Illegal expression" exp)))))))



