;*=====================================================================*/
;*    serrano/prgm/project/skribe/tools/skribeinfo/main.scm            */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Fri Oct 12 14:57:58 2001                          */
;*    Last change :  Mon Sep 15 09:49:45 2003 (serrano)                */
;*    Copyright   :  2001-03 Manuel Serrano                            */
;*    -------------------------------------------------------------    */
;*    The entry point of the info->skribe translator                   */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module skribeinfo
   (import __skribeinfo_parser
	   __skribeinfo_lexer
	   __skribeinfo_engine
	   __skribeinfo_prelude
	   __skribeinfo_param)
   (main   main))

;*---------------------------------------------------------------------*/
;*    main ...                                                         */
;*---------------------------------------------------------------------*/
(define (main argv)
   (define (usage args-parse-usage)
      (print "usage: skribeinfo [options] [input]")
      (newline)
      (args-parse-usage #f))
   (let ((stage 'scr)
	 (dest #f)
	 (in #f))
      (args-parse (cdr argv)
	 ((("-h" "--help") (help "This help message"))
	  (usage args-parse-usage)
	  (exit 0))
	 ((("--options") (help "Display the options and exit"))
	  (args-parse-usage #t)
	  (exit 0))
	 (("--sexp" (help "Produce only s-expressions"))
	  (set! stage 'sexp))
	 (("-o" ?out (help "Set the destination file"))
	  (set! dest out))
	 (("--sui" ?info ?sui (help "Associate a SUI file with an INFO file"))
	  (set! *info-sui* (cons (cons info sui) *info-sui*)))
	 (("-v" (help "Verbose mode"))
	  (set! *skribeinfo-verbose* (+ 1 *skribeinfo-verbose*)))
	 ((("-s" "--style") ?style (help "Add the style file"))
	  (set! *skribeinfo-styles* (cons style *skribeinfo-verbose*)))
	 ((("-r" "--reset-style") ?style (help "Reset the style files list"))
	  (set! *skribeinfo-styles* '()))
	 (else
	  (set! in else)))
      (if (string? dest)
	  (with-output-to-file dest (lambda () (skribeinfo in stage)))
	  (skribeinfo in stage))))

;*---------------------------------------------------------------------*/
;*    skribeinfo ...                                                   */
;*---------------------------------------------------------------------*/
(define (skribeinfo in stage)
   (let* ((port (if (string? in)
		    (open-input-file in)
		    (current-input-port)))
	  (sexp (parse-port port)))
      (if (eq? stage 'sexp)
	  (begin
	     (write sexp)
	     (newline))
	  (begin
	     (print ";; styles")
	     (for-each (lambda (f)
			  (print "(skribe-load \"" f "\")"))
		       *skribeinfo-styles*)
	     (newline)
	     (print ";; indexes")
	     (for-each (lambda (i)
			  (print `(define ,i
				     (make-index ,(format "\"~a\"" i)))))
		       *indexes*)
	     (newline)
	     (print ";; prelude")
	     (for-each (lambda (e)
			  (write e)
			  (newline))
		       (skribeinfo-prelude))
	     (newline)
	     (print ";; body")
	     (write (texi->skribe sexp))
	     (newline)))))

