;*=====================================================================*/
;*    serrano/prgm/project/skribe/tools/skribeinfo/parser.scm          */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Thu Oct 11 09:30:45 2001                          */
;*    Last change :  Fri Oct 24 15:07:56 2003 (serrano)                */
;*    Copyright   :  2001-03 Manuel Serrano                            */
;*    -------------------------------------------------------------    */
;*    The parser for texinfo format.                                   */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __skribeinfo_parser
   (import __skribeinfo_lexer
	   __skribeinfo_engine
	   __skribeinfo_param)
   (export (parse-file ::bstring)
	   (parse-string ::bstring . coord)
	   (parse-port ::input-port)))

;*---------------------------------------------------------------------*/
;*    parse-file ...                                                   */
;*---------------------------------------------------------------------*/
(define (parse-file file)
   (if (> *skribeinfo-verbose* 0)
       (fprint (current-error-port) "  [including: " file "]"))
   (let ((p (open-input-file file)))
      (if (not (input-port? p))
	  (error "parse-file" "Can't open file for input" file)
	  (unwind-protect (parse-port p)
			  (close-input-port p)))))
 
;*---------------------------------------------------------------------*/
;*    parse-string ...                                                 */
;*---------------------------------------------------------------------*/
(define (parse-string string . coord)
   (with-input-from-string string
      (lambda ()
	 (if (pair? coord)
	     (try (read/lalrp parser lexer (current-input-port))
		  (lambda (escape proc mes obj)
		     (match-case obj
			((?token (?fname . ?pos) . ?-)
			 (error/location proc "parse error" token
					 (caar coord) (+fx (cdar coord) pos)))
		 	(else
			 (error/location proc mes obj
					 (caar coord) (cdar coord))))))
	     (read/lalrp parser lexer (current-input-port))))))
   
;*---------------------------------------------------------------------*/
;*    parse-port ...                                                   */
;*---------------------------------------------------------------------*/
(define (parse-port port)
   (try (read/lalrp parser lexer port)
	(lambda (escape proc mes obj)
	   (match-case obj
	      ((?token (?fname . ?pos) . ?-)
	       (error/location proc "parse error" token fname pos))
	      (else
	       (notify-error proc mes obj)
	       (error proc mes obj))))))

;*---------------------------------------------------------------------*/
;*    *node-mark* ...                                                  */
;*---------------------------------------------------------------------*/
(define *node-mark* #f)

;*---------------------------------------------------------------------*/
;*    parser ...                                                       */
;*---------------------------------------------------------------------*/
(define parser
   
   (lalr-grammar
      
      ;; tokens
      (STRING
       LINEBREAK
       BRACLO
       BRAOPEN
       NODE
       B I R W T EMPH DFN SAMP SC STRONG
       KBD KEY
       CODE VAR
       REF XREF PXREF INFOREF
       COPYRIGHT EXPANSION EQUIV ERROR RESULT PRINT DOTS FILE FOOTNOTE MATH TEX
       POINT MINUS
       EMAIL URL CITE ANCHOR NEED
       IGNORE |END IGNORE|
       TABLE |TABLE CODE| |TABLE BULLET| |TABLE EMPH| |TABLE SAMP|
       |TABLE ASIS| |TABLE T| |TABLE R| |TABLE MINUS| |TABLE VAR| |TABLE DFN|
       |TABLE KBD| |TABLE B| |END TABLE|
       ITEMIZE |ITEMIZE BULLET| |ITEMIZE MINUS| |END ITEMIZE|
       ENUMERATE |END ENUMERATE|
       ITEM
       ITEMX
       IFSET |END IFSET|
       MENU |END MENU|
       SETFILENAME SETTITLE
       PARAGRAPHINDENT
       FOOTNOTESTYLE
       FINALOUT
       SET VALUE
       INCLUDE
       SETCHAPTERNEWPAGE
       SHORTTITLEPAGE
       TITLEPAGE |END TITLEPAGE|
       TITLE SUBTITLE
       ATITLE ASUBTITLE
       AUTHOR AAUTHOR
       PAGE
       VSKIP
       DIRCATEGORY
       DIRENTRY |END DIRENTRY|
       SMALLLISP |END SMALLLISP|
       LISP |END LISP|
       FLUSHLEFT |END FLUSHLEFT|
       FLUSHRIGHT |END FLUSHRIGHT|
       SMALLEXAMPLE |END SMALLEXAMPLE|
       EXAMPLE |END EXAMPLE|
       GROUP |END GROUP|
       FORMAT |END FORMAT|
       DISPLAY |END DISPLAY|
       QUOTATION |END QUOTATION|
       CHAPTER SECTION SUBSECTION SUBSUBSECTION
       UNNUMBERED UNNUMBEREDSEC UNNUMBEREDSUBSEC UNNUMBEREDSUBSUBSEC
       CINDEX VINDEX KINDEX FINDEX TINDEX PINDEX
       PRINTINDEX SYNCODEINDEX SYNINDEX
       REFILL NOINDENT EXDENT
       CONTENTS SUMMARYCONTENTS
       BYE
       DEFN |END DEFN|
       DEFFN |END DEFFN|
       DEFFNX
       DEFTP |END DEFTP|
       DEFMETHOD |END DEFMETHOD|
       DEFUN DEFUNX |END DEFUN|
       DEFMAC DEFMACX |END DEFMAC|
       DEFVAR DEFVARX |END DEFVAR|
       DEFVR DEFVRX |END DEFVR|
       DEFOPT DEFOPTX |END DEFOPT|
       DEFSPEC DEFSPECX |END DEFSPEC|
       SP
       CENTER
       SMALLBOOK)
      
      ;; commands
      (commands
       (() '())
       ((command commands) (cons command commands)))
      
      ;; command
      (command
       ((ignore) '())
       ((text) text)
       ((argument-command) argument-command)
       ((line-command) line-command)
       ((BRAOPEN commands BRACLO) `(list ,@commands)))
      
      ;; ignore
      (ignore
       ((MENU commands |END MENU|) '())
       ((SETFILENAME) '())
       ((FOOTNOTESTYLE) '())
       ((PARAGRAPHINDENT) '())
       ((FINALOUT) '())
       ((SETCHAPTERNEWPAGE) '())
       ((SHORTTITLEPAGE) '())
       ((PAGE) '())
       ((DIRCATEGORY) '())
       ((DIRENTRY commands |END DIRENTRY|) '())
       ((VSKIP) '())
       ((REFILL) '())
       ((CONTENTS) '())
       ((BYE) '())
       ((SMALLBOOK) '())
       ((NEED) '()))
      
      ;; text
      (text
       ((STRING)
	(cadr STRING))
       ((LINEBREAK)
	`(linebreak ,(cadr LINEBREAK))))
      
      ;; argument-command
      (argument-command
       ((acommand BRACLO)
	acommand))
      
      (acommand
       ((VALUE STRING)
	`(value ,(cadr STRING)))
       ((COPYRIGHT)
	'(symbol "copyright"))
       ((MINUS)
	"-")
       ((POINT)
	"-!-")
       ((TEX)
	"TeX")
       ((DOTS)
	"...")
       ((B commands)
	(if (pair? commands)
	    `(bold ,@commands)
	    ""))
       ((ATITLE commands)
	(if (pair? commands)
	    `(title ,@commands)
	    ""))
       ((ASUBTITLE commands)
	(if (pair? commands)
	    `(subtitle ,@commands)
	    ""))
       ((AAUTHOR commands)
	`(author ,@commands))
       ((I commands)
	(if (pair? commands)
	    `(it ,@commands)
	    ""))
       ((R commands)
	`(list ,@commands))
       ((EMAIL commands)
	(mailto commands))
       ((URL commands)
	`(ref :url (list ,@commands)))
       ((W commands)
	(if (pair? commands)
	    `(bold ,@commands)
	    ""))
       ((T commands)
	(if (pair? commands)
	    `(tt ,@commands)
	    ""))
       ((CODE commands)
	(if (pair? commands)
	    `(code ,@commands)
	    ""))
       ((VAR commands)
	(if (pair? commands)
	    `(code ,@commands)
	    ""))
       ((EMPH commands)
	(if (pair? commands)
	    `(emph ,@commands)
	    ""))
       ((DFN commands)
	(string-append "\"" (stringify commands) "\""))
       ((SAMP commands)
	(if (pair? commands)
	    `(samp ,@commands)
	    ""))
       ((SC commands)
	(if (pair? commands)
	    `(sc ,@commands)
	    ""))
       ((STRONG commands)
	(if (pair? commands)
	    `(bold ,@commands)
	    ""))
       ((KBD commands)
	(if (pair? commands)
	    `(kbd ,@commands)
	    ""))
       ((KEY commands)
	(if (pair? commands)
	    `(kbd ,@commands)
	    ""))
       ((REF commands)
	(ref '@ref REF commands))
       ((XREF commands)
	`(list "See " ,(ref '@xref XREF commands)))
       ((INFOREF commands)
	(inforef commands))
       ((PXREF commands)
	`(list "see " ,(ref '@pxref PXREF commands)))
       ((EXPANSION)
	"==>")
       ((EQUIV)
	"<=>")
       ((ERROR)
	"error-->")
       ((FOOTNOTE commands)
	`(footnote :note ,@commands))
       ((MATH commands)
	(math commands))
       ((CITE commands)
	`(it (list ,@commands)))
       ((ANCHOR text)
	`(mark ,text))
       ((RESULT)
	"=>")
       ((RESULT command)
	`(list "=>" ,command))
       ((PRINT)
	"-|")
       ((FILE commands)
	`(list "`" ,@commands "'")))
      
      ;; line-command
      (line-command
       ((NODE)
	(set! *node-mark*
	      `(mark ,(string-skip (nthword (car (cadr NODE)) 0 #\,)
				    #\space #\tab))))
       ((INCLUDE)
	`(list ,@(parse-file (car (cadr INCLUDE)))))
       ((itemize)
	itemize)
       ((table)
	table)
       ((ITEM)
	(if (pair? (cadr ITEM))
	    `(item ,@(cadr ITEM) "\n")
	    `(item)))
       ((ITEMX)
	`(itemx ,@(cadr ITEMX) "\n"))
       ((SET)
	(parse-set (cadr SET)))
       ((CHAPTER)
	(if *node-mark*
	    (let ((lbl *node-mark*))
	       (set! *node-mark* #f)
	       `(chapter :title ,@(cadr CHAPTER) ,lbl))
	    `(chapter :title ,@(cadr CHAPTER))))
       ((SECTION)
	(if *node-mark*
	    (let ((lbl *node-mark*))
	       (set! *node-mark* #f)
	       `(section :title (list ,@(cadr SECTION)) ,lbl))
	    `(section :title (list ,@(cadr SECTION)))))
       ((SUBSECTION)
	(if *node-mark*
	    (let ((lbl *node-mark*))
	       (set! *node-mark* #f)
	       `(subsection :title ,@(cadr SUBSECTION) ,lbl))
	    `(subsection :title ,@(cadr SUBSECTION))))
       ((SUBSUBSECTION)
	(if *node-mark*
	    (let ((lbl *node-mark*))
	       (set! *node-mark* #f)
	       `(subsubsection :title ,@(cadr SUBSUBSECTION) ,lbl))
	    `(subsubsection :title ,@(cadr SUBSUBSECTION))))
       ((UNNUMBERED)
	(if *node-mark*
	    (let ((lbl *node-mark*))
	       (set! *node-mark* #f)
	       `(chapter :number #f :toc #t :file #f :title ,@(cadr UNNUMBERED) ,lbl))
	    `(chapter :number #f :toc #t :file #f :title ,@(cadr UNNUMBERED))))
       ((UNNUMBEREDSEC)
	(if *node-mark*
	    (let ((lbl *node-mark*))
	       (set! *node-mark* #f)
	       `(section :number #f :title ,@(cadr UNNUMBEREDSEC) ,lbl))
	    `(section :number #f :title ,@(cadr UNNUMBEREDSEC))))
       ((UNNUMBEREDSUBSEC)
	(if *node-mark*
	    (let ((lbl *node-mark*))
	       (set! *node-mark* #f)
	       `(subsection :number #f :title ,@(cadr UNNUMBEREDSUBSEC) ,lbl))
	    `(subsection :number #f :title ,@(cadr UNNUMBEREDSUBSEC))))
       ((UNNUMBEREDSUBSUBSEC)
	(if *node-mark*
	    (let ((lbl *node-mark*))
	       (set! *node-mark* #f)
	       `(subsubsection :number #f :title ,@(cadr UNNUMBEREDSUBSUBSEC) ,lbl))
	    `(subsubsection :number #f :title ,@(cadr UNNUMBEREDSUBSUBSEC))))
       ((SETTITLE)
	`(header ,@(cadr SETTITLE)))
       ((TITLEPAGE commands |END TITLEPAGE|)
	`(title ,@commands))
       ((TITLE)
	`(title ,@(cadr TITLE)))
       ((SUBTITLE) 
	`(subtitle ,@(cadr SUBTITLE)))
       ((AUTHOR)
	`(author ,@(cadr AUTHOR)))
       ((IGNORE commands |END IGNORE|)
	(marks-only commands))
       ((IFSET commands |END IFSET|)
	(if (variable-bound? (car (cadr IFSET)))
	    `(list ,@commands)
	    (marks-only commands)))
       ((SMALLLISP commands |END SMALLLISP|)
	(prgm 'bigloo '*skribeinfo-prgm-color* commands))
       ((LISP commands |END LISP|)
	(prgm 'bigloo '*skribeinfo-prgm-color* commands))
       ((SMALLEXAMPLE commands |END SMALLEXAMPLE|)
	(prgm #f '*skribeinfo-example-color* commands))
       ((EXAMPLE commands |END EXAMPLE|)
	(prgm #f '*skribeinfo-example-color* commands))
       ((GROUP commands |END GROUP|)
	`(list ,@commands))
       ((FORMAT commands |END FORMAT|)
	(prgm #f '*skribeinfo-example-color* commands))
       ((DISPLAY commands |END DISPLAY|)
	(prgm #f '*skribeinfo-display-color* commands))
       ((QUOTATION commands |END QUOTATION|)
	`(font :size "=-1" ,@commands))
       ((DEFN commands |END DEFN|)
	(deffn (cadr DEFN) commands))
       ((DEFFN commands |END DEFFN|)
	(deffn (cadr DEFFN) commands))
       ((DEFFNX)
	`(deffnx ,@(cadr DEFFNX)))
       ((DEFTP commands |END DEFTP|)
	(deffn (cadr DEFTP) commands))
       ((DEFUN commands |END DEFUN|)
	(defun (cadr DEFUN) commands))
       ((DEFUNX)
	`(generic-defx ,@(cadr DEFUNX)))
       ((DEFMAC commands |END DEFMAC|)
	(defmac (cadr DEFMAC) commands))
       ((DEFMACX)
	`(generic-defx ,@(cadr DEFMACX)))
       ((DEFVAR commands |END DEFVAR|)
	(defvar (cadr DEFVAR) commands))
       ((DEFVARX)
	`(generic-defx ,@(cadr DEFVARX)))
       ((DEFVR commands |END DEFVR|)
	(defvar (cadr DEFVR) commands))
       ((DEFVRX)
	`(generic-defx ,@(cadr DEFVRX)))
       ((DEFOPT commands |END DEFOPT|)
	(defopt (cadr DEFOPT) commands))
       ((DEFOPTX)
	`(generic-defx ,@(cadr DEFOPTX)))
       ((DEFSPEC commands |END DEFSPEC|)
	(defspec (cadr DEFSPEC) commands))
       ((DEFSPECX)
	`(generic-defx ,@(cadr DEFSPECX)))
       ((DEFMETHOD commands |END DEFMETHOD|)
	(defmethod (cadr DEFMETHOD) commands))
       ((SYNCODEINDEX)
	(let ((w1 (nthword (car (cadr SYNCODEINDEX)) 0))
	      (w2 (nthword (car (cadr SYNCODEINDEX)) 1)))
	   (syncodeindex (string->symbol (format "*~a*" w1))
			 (string->symbol (format "*~a*" w2)))))
       ((SYNINDEX)
	(let ((w1 (nthword (car (cadr SYNINDEX)) 0))
	      (w2 (nthword (car (cadr SYNINDEX)) 1)))
	   (synindex (string->symbol (format "*~a*" w1))
		     (string->symbol (format "*~a*" w2)))))
       ((FLUSHLEFT commands |END FLUSHLEFT|)
	`(flush :side 'left ,@commands))
       ((FLUSHRIGHT commands |END FLUSHRIGHT|)
	`(flush :side 'left ,@commands))
       ((CINDEX)
	(index-entry '*cp* (cadr CINDEX)))
       ((PINDEX)
	(index-entry '*pg* (cadr PINDEX)))
       ((VINDEX)
	(index-entry '*vr* (cadr VINDEX)))
       ((KINDEX)
	(index-entry '*ky* (cadr KINDEX)))
       ((FINDEX)
	(index-entry '*fn* (cadr FINDEX)))
       ((TINDEX)
	(index-entry '*tp* (cadr TINDEX)))
       ((PRINTINDEX)
	(printindex (string->symbol (format "*~a*" (car (cadr PRINTINDEX))))))
       ((SUMMARYCONTENTS)
	`(toc))
       ((NOINDENT)
	(if (pair? (cadr NOINDENT))
	    `(list ,@(cadr NOINDENT))
	    '()))
       ((EXDENT)
	(if (pair? (cadr EXDENT))
	    `(list ,@(cadr EXDENT))
	    '()))
       ((SP)
	(if (and (pair? (cdr SP)) (pair? (cadr SP)))
	    (let ((n (string->integer (car (cadr SP)))))
	       `(linebreak ,n))
	    '(linebreak 1)))
       ((CENTER)
	(if (pair? (cadr CENTER))
	    `(center ,@(cadr CENTER))
	    '())))
      
      ;; itemize
      (itemize
       ((ITEMIZE commands |END ITEMIZE|)
	(itemize 'itemize commands))
       ((|ITEMIZE BULLET| commands |END ITEMIZE|)
	(itemize 'itemize commands))
       ((|ITEMIZE MINUS| commands |END ITEMIZE|)
	(itemize 'itemize commands))
       ((ENUMERATE commands |END ENUMERATE|)
	(itemize 'enumerate commands)))
      
      ;; table
      (table
       ((TABLE commands |END TABLE|)
	(table commands (lambda (x) x)))
       ((|TABLE CODE| commands |END TABLE|)
	(table commands (lambda (x) `(code ,x))))
       ((|TABLE BULLET| commands |END TABLE|)
	(table commands (lambda (x) x)))
       ((|TABLE EMPH| commands |END TABLE|) 
	(table commands (lambda (x) `(emph ,x))))
       ((|TABLE SAMP| commands |END TABLE|) 
	(table commands (lambda (x) `(samp ,x))))
       ((|TABLE ASIS| commands |END TABLE|) 
	(table commands (lambda (x) x)))
       ((|TABLE T| commands |END TABLE|) 
	(table commands (lambda (x) `(tt ,x))))
       ((|TABLE VAR| commands |END TABLE|) 
	(table commands (lambda (x) `(code ,x))))
       ((|TABLE DFN| commands |END TABLE|) 
	(table commands (lambda (x) `(code ,x))))
       ((|TABLE KBD| commands |END TABLE|) 
	(table commands (lambda (x) `(kbd ,x))))
       ((|TABLE B| commands |END TABLE|) 
	(table commands (lambda (x) `(bold ,x))))
       ((|TABLE R| commands |END TABLE|) 
	(table commands (lambda (x) x)))
       ((|TABLE MINUS| commands |END TABLE|) 
	(table commands (lambda (x) x))))))

;*---------------------------------------------------------------------*/
;*    stringify ...                                                    */
;*---------------------------------------------------------------------*/
(define (stringify o)
   (cond
      ((string? o)
       o)
      ((pair? o)
       (match-case o
	  ((value . ?-)
	   o)
	  (else
	   (if (null? (cdr o))
	       (stringify (car o))
	       (apply string-append (map stringify o))))))
      ((symbol? o)
       "")
      (else
       (error "stringify" "Illegal value" o))))

;*---------------------------------------------------------------------*/
;*    marks-only ...                                                   */
;*---------------------------------------------------------------------*/
(define (marks-only lst)
   (define (find-marks expr res)
      (if (not (pair? expr))
	  res
	  (match-case expr
	     ((mark . ?-)
	      (cons expr res))
	     (else
	      (find-marks (car expr) (find-marks (cdr expr) res))))))
   (let loop ((lst lst)
	      (res '()))
      (if (null? lst)
	  (if (pair? res)
	      `(list ,@res)
	      '())
	  (loop (cdr lst)
		(find-marks (car lst) res)))))

;*---------------------------------------------------------------------*/
;*    parse-set ...                                                    */
;*---------------------------------------------------------------------*/
(define (parse-set token)
   (let* ((str (car token))
	  (var (nthword str 0))
	  (val (if (<fx (string-length var) (-fx (string-length str) 1))
		   (let ((val0 (substring str
					  (+fx (string-length var) 1)
					  (string-length str))))
		      (cons val0 (cdr token)))
		   (cdr token))))
      (bind-variable var val)
      '()))

;*---------------------------------------------------------------------*/
;*    mailto ...                                                       */
;*---------------------------------------------------------------------*/
(define (mailto commands)
   (let ((email (apply string-append (map (lambda (x)
					     (if (string? x)
						 x
						 ""))
					  commands))))
      (if (>fx (string-length email) 0)
	  `(mailto :email ,email ,@commands)
	  `(mailto ,@commands))))

;*---------------------------------------------------------------------*/
;*    table ...                                                        */
;*---------------------------------------------------------------------*/
(define (table commands deco)
   (define (table-lignes commands)
      (define (table-first-column commands lines)
	 (let loop ((commands commands)
		    (items '()))
	    (if (null? commands)
		(reverse! lines)
		(match-case (car commands)
		   ((item . ?item)
		    (if (not (null? items))
			;; this should be an error but it turns out that
			;; texinfo is laxist
			(loop (cdr commands)
			      (append (reverse! (map deco item)) items))
			(loop (cdr commands)
			      (reverse! (map deco item)))))
		   ((itemx . ?item)
		    (if (null? items)
			(error "table" "Illegal itemx" (car commands))
			(loop (cdr commands)
			      (append (reverse! (map deco item))
				      items))))
		   (else
		    (cond
		       ((null? items)
			(loop (cdr commands) items))
		       (else
			(table-second-column commands
					     (reverse! items)
					     lines))))))))
      (define (table-second-column commands dd lines)
	 (let loop ((commands commands)
		    (vals '()))
	    (if (null? commands)
		(table-first-column '()
				    (cons `(item :key (list ,@dd)
						 ,@(reverse! vals))
					  lines))
		(match-case (car commands)
		   ((item . ?item)
		    (table-first-column commands
					(cons `(item :key (list ,@dd)
						     ,@(reverse! vals))
					      lines)))
		   (else
		    (loop (cdr commands) (cons (car commands) vals)))))))
      (table-first-column commands '()))
   `(description ,@(table-lignes commands)))
     
;*---------------------------------------------------------------------*/
;*    itemize ...                                                      */
;*---------------------------------------------------------------------*/
(define (itemize markup commands)
   (define (itemize-lignes commands)
      (define (itemize-item commands lines)
	 (let loop ((commands commands)
		    (items '()))
	    (if (null? commands)
		(reverse! lines)
		(match-case (car commands)
		   ((item . ?item)
		    (itemize-entry (cdr commands) (reverse! item) lines))
		   (else
		    (loop (cdr commands) items))))))
      (define (itemize-entry commands vals lines)
	 (let loop ((commands commands)
		    (vals vals))
	    (if (null? commands)
		(itemize-item '() (cons `(item ,@(reverse! vals)) lines))
		(match-case (car commands)
		   ((item . ?item)
		    (itemize-item commands
				  (cons `(item ,@(reverse! vals)) lines)))
		   ((itemx . ?item)
		    (error "itemize" "Illegal itemx" (car commands)))
		   (else
		    (loop (cdr commands) (cons (car commands) vals)))))))
      (itemize-item commands '()))
   `(,markup ,@(itemize-lignes commands)))

;*---------------------------------------------------------------------*/
;*    *deffn-marks* ...                                                */
;*---------------------------------------------------------------------*/
(define *deffn-marks* '())

;*---------------------------------------------------------------------*/
;*    deffn ...                                                        */
;*---------------------------------------------------------------------*/
(define (deffn def commands)
   (define (make-kind name rest kind)
      (if (string? name)
	  (if (and *auto-deffn-index* (not (member name *deffn-marks*)))
	      (begin
		 (set! *deffn-marks* (cons name *deffn-marks*))
		 `(list (index :index *pg* ,name :shape (code ,name))
			(mark ,name)
			(table :width 100.
			       (tr (td (bold ,name) :align 'left (it ,@rest))
				   (td :align 'right ,@kind)))))
	      `(list (table :width 100.
			    (tr (td (bold ,name) :align 'left (it ,@rest))
				(td :align 'right ,@kind)))))
	  `(list (table :width 100.
			(tr (td :align 'left (it ,@rest))
			    (td :align 'right ,@kind))))))
   (define (make-deffn-title title)
      (match-case title
	 (((list . ?kind) (and ?def (? string?)))
	  (let* ((name (nthword def 0))
		 (rest (substring def
				  (+fx 1 (string-length name))
				  (string-length def))))
	     (make-kind name (list rest) kind)))
	 (((list . ?kind) (and ?def (? string?)) . ?rests)
	  (let ((name (nthword def 0)))
	     (if (not name)
		 (make-kind #f rests kind)
		 (let ((rest (substring def
					(+fx 1 (string-length name))
					(string-length def))))
		    (make-kind name (cons rest rests) kind)))))
	 (((list . ?kind) . ?rest)
	  (make-kind #f rest kind))
	 (else
	  title)))
   `(list (frame :width 100.
		 :margin 10
		 ,(make-deffn-title def)
		 ,@(map (lambda (x)
			   (match-case x
			      ((deffnx . ?rest)
			       (make-deffn-title rest))
			      (else
			       x)))
			commands))
	  (linebreak 1)))

;*---------------------------------------------------------------------*/
;*    defun ...                                                        */
;*---------------------------------------------------------------------*/
(define (defun def commands)
   (generic-def def commands "function"))

;*---------------------------------------------------------------------*/
;*    defmethod ...                                                    */
;*---------------------------------------------------------------------*/
(define (defmethod def commands)
   (generic-def def commands "method"))

;*---------------------------------------------------------------------*/
;*    defmac ...                                                       */
;*---------------------------------------------------------------------*/
(define (defmac def commands)
   (generic-def def commands "macro"))

;*---------------------------------------------------------------------*/
;*    defvar ...                                                       */
;*---------------------------------------------------------------------*/
(define (defvar def commands)
   (generic-def def commands "variable"))

;*---------------------------------------------------------------------*/
;*    defopt ...                                                       */
;*---------------------------------------------------------------------*/
(define (defopt def commands)
   (generic-def def commands "user option"))

;*---------------------------------------------------------------------*/
;*    defspec ...                                                      */
;*---------------------------------------------------------------------*/
(define (defspec def commands)
   (generic-def def commands "special form"))

;*---------------------------------------------------------------------*/
;*    generic-def ...                                                  */
;*---------------------------------------------------------------------*/
(define (generic-def def commands kind)
   (define (make-def rest)
      `(list (table :width 100.
		    (tr (td (it ,rest))
			(td :align 'right ,kind)))))
   (define (make-def-title title)
      (make-def (stringify title)))
   `(list (frame :width 100.
		 :margin 10
		 ,(make-def-title def)
		 ,@(map (lambda (x)
			   (match-case x
			      ((generic-defx . ?rest)
			       (make-def-title x))
			      (else
			       x)))
			commands))
	  (linebreak 1)))

;*---------------------------------------------------------------------*/
;*    ref ...                                                          */
;*---------------------------------------------------------------------*/
(define (ref cmd REF commands)
   (let* ((str (apply string-append commands))
	  (unspace (lambda (s)
			  (if (string? s)
			      (string-skip s #\space)
			      s)))
	  (w0 (unspace (nthword2 str 0 #\,)))
	  (w1 (unspace (nthword2 str 1 #\,)))
	  (w2 (unspace (nthword2 str 2 #\,)))
	  (w3 (nthword2 str 3 #\,)))
      (cond
	 ((not (string? w0))
	  (error "@ref" "Illegal reference" `(,cmd ,(car REF))))
	 ((not (or w1 w2 w3))
	  ;; plain reference to a local node
	  `(ref :mark ,(string-replace w0 #\Newline #\Space)))
	 ((and (string? w2) (assoc w2 *info-sui*))
	  ;; a reference to another Skribe document
	  (let ((cell (cdr (assoc w2 *info-sui*)))
		(body (cond
			 ((or (pair? w1)
			      (and (string? w1) (>fx (string-length w1) 0)))
			  w1)
			 ((and (string? w0) (>fx (string-length w0) 0))
			  (string-append (prefix w2) ", " w0))
			 ((pair? w0)
			  (cons (string-append (prefix w2) ", ") w0))
			 (else
			  (prefix w2)))))
	     (if (and (string? w0) (>fx (string-length w0) 0))
		 `(ref :skribe ,cell :mark ,w0 :text ,body)
		 `(ref :skribe ,cell :text ,body))))
	 ((string? w3)
	  ;; a pure tex reference
	  (let* ((base (string-append "info-file `" w3 "'"))
 		 (ebase (string-append base ", " (if w2 w2 w0))))
	     (string-replace ebase #\Newline #\Space)))
	 (else
	  ;; a reference with a customized text
	  (let* ((user-title1 (if w2 w2 w0))
		 (user-title2 (if (and (string? w1)
				       (>fx (string-length w1) 0))
				  (string-append "(" w1 "): " user-title1)
				  user-title1)))
	     `(ref :mark
		   ,(string-replace w0 #\Newline #\Space)
		   :text
		   ,(string-replace user-title2 #\Newline #\Space)))))))

;*---------------------------------------------------------------------*/
;*    inforef ...                                                      */
;*---------------------------------------------------------------------*/
(define (inforef commands)
   (let* ((str (apply string-append commands))
	  (w0 (nthword str 0 #\,))
	  (w1 (nthword str 1 #\,))
	  (w2 (nthword str 2 #\,)))
      (cond
	 ((not (string? w0))
	  (error "@ref" "Illegal reference" 'inforef))
	 ((not (or w1 w2))
	  ;; plain reference to a local node
	  `(ref :mark ,(string-replace w0 #\Newline #\Space)))
	 ((string? w2)
	  ;; a pure tex reference
	  (let* ((base (string-append "info-file `" w2 "'"))
		 (ebase (string-append base ", " (if w2 w2 w0))))
	     (string-replace ebase #\Newline #\Space)))
	 (else
	  ;; a reference with a customized text
	  (let* ((user-title1 (if w2 w2 w0))
		 (user-title2 (if (and (string? w1)
				       (>fx (string-length (string-skip w1 #\Space)) 0))
				  (string-append "(" w1 "): " user-title1)
				  user-title1)))
	     `(ref :mark
		   ,(string-replace w0 #\Newline #\Space)
		   :text
		   ,(string-replace user-title2 #\Newline #\Space)))))))

;*---------------------------------------------------------------------*/
;*    math ...                                                         */
;*    -------------------------------------------------------------    */
;*    The math handling is extremelly weak. We only take care to       */
;*    _ and ^ that are replaced with (sub ...) and (sup ...).          */
;*---------------------------------------------------------------------*/
(define (math cmds)
   (let* ((str (apply string-append cmds))
	  (len (string-length str))
	  (len-1 (-fx len 1)))
      (let loop ((s 0)
		 (r 0)
		 (acc '()))
	 (if (>=fx r len-1)
	     (let ((l (reverse!
		       (if (=fx s len) acc (cons (substring str s len) acc)))))
		(cond
		   ((null? l)
		    "")
		   ((null? (cdr l))
		    (car l))
		   (else
		    `(list ,@l))))
	     (let ((c (string-ref str r)))
		(case c
		   ((#\_ #\^)
		    (let ((sub `(,(if (char=? c #\_) 'sub 'sup)
				 ,(string (string-ref str (+fx r 1))))))
		       (if (=fx s r)
			   (loop (+fx r 2)
				 (+fx r 2)
				 (cons sub acc))
			   (loop (+fx r 2)
				 (+fx r 2)
				 (cons* sub (substring str s r) acc)))))
		   (else
		    (loop s (+fx r 1) acc))))))))
      
;*---------------------------------------------------------------------*/
;*    prgm ...                                                         */
;*---------------------------------------------------------------------*/
(define (prgm lg bg body)
   `(center
     (color :bg ,bg :width *skribeinfo-prgm-width*
	    (prog :line #f (source :language ,lg ,@body)))))
	   
